"""Very thin wrapper around the Strava API v3

    STRAVA_ACCESS_TOKEN must be explicitly provided as an input

    ALL returned values are python objects e.g. dict or list
"""
import requests
import logging

logger = logging.getLogger(__name__)


def retrieve_athlete(access_token):
    """Retrieve current(authenticated) athlete

    API V3: https://strava.github.io/api/v3/athlete/#get-details

    Parameters
    ----------
    access_token : str
        Settings/My API Applications/Your Access Token

    Returns
    -------
    dict
    """

    endpoint_url = "https://www.strava.com/api/v3/athlete"

    r = requests.get(endpoint_url, headers=authorization_header(access_token))

    r.raise_for_status()
    athlete = r.json()

    return athlete


def retrieve_zones(access_token, **kwargs):
    """Retrieve Power and Heartrate zones

    API V3: https://strava.github.io/api/v3/athlete/#zones

    Parameters
    ----------
    access_token : str
        Settings/My API Applications/Your Access Token

    Returns
    -------
    dict
    """

    endpoint_url = "https://www.strava.com/api/v3/athlete/zones"

    r = requests.get(endpoint_url, headers=authorization_header(access_token))

    r.raise_for_status()
    zones = r.json()

    return zones


def retrieve_activity(activity_id, access_token):
    """Retrieve a detailed representation of activity

    API V3: https://strava.github.io/api/v3/activities/#get-details

    Parameters
    ----------
    activity_id: int
    access_token : str
        Settings/My API Applications/Your Access Token

    Returns
    -------
    dict
    """

    endpoint_url = "https://www.strava.com/api/v3/activities/{}".format(activity_id)

    r = requests.get(endpoint_url, headers=authorization_header(access_token))

    r.raise_for_status()
    activity = r.json()

    return activity


def retrieve_streams(activity_id, access_token, **kwargs):
    """Retrieve activity streams

    API V3: https://strava.github.io/api/v3/streams/#activity

    Parameters
    ----------
    activity_id : int
    access_token : str
        Settings/My API Applications/Your Access Token
    type: {None, 'original'}
        Returns serialized original API response if set to 'original'

    Returns
    -------
    streams : list
        Streams are the list of dicts
    """

    types = kwargs.get(
        "types",
        "time,latlng,distance,altitude,velocity_smooth,heartrate,cadence,watts,temp,moving,grade_smooth",
    )

    endpoint_url = "https://www.strava.com/api/v3/activities/{}/streams/{}".format(
        activity_id, types
    )

    r = requests.get(endpoint_url, headers=authorization_header(access_token))

    r.raise_for_status()
    streams = r.json()

    if streams and not kwargs.get("type", None):

        streams = stream2dict(streams)

    return streams


def stream2dict(stream_list):
    """Convert stream list into stream dict

    Parameters
    ----------
    stream_list : list
        Stream in list form (list of dicts), as returned by Strava API v3

    Returns
    -------
    stream_dict : dict
        Stream in dict form, with key set to *stream name* and value set to the actual stream list.
        In this form, the stream is ready to be consumed by pandas
    """

    stream_dict = {}

    for s in stream_list:

        stream_dict.update({s["type"]: s["data"]})

    return stream_dict


def zones2list(zones, type="power"):
    """Convert zones Strava response into a list

    Parameters
    ----------
    zones : dict
        Strava API zones response
    type : {"power", "heart_rate"}

    Returns
    -------
    y : list
        Zones boundaries with left edge set to -1 and right to 10000
    """

    y = [x["min"] for x in zones[type]["zones"]]
    y[0] = -1
    y.append(10000)

    return y


def authorization_header(access_token):
    """Authorization header dict to be used with requests.get()

    Parameters
    ----------
    access_token : str
        Settings/My API Applications/Your Access Token

    Returns
    -------
    header : dict
    """

    rv = {"Authorization": "Bearer {}".format(access_token)}

    return rv



from goldencheetahlib.client import GoldenCheetahClient

from .models.dataframes import WorkoutDataFrame


class Client(GoldenCheetahClient):
    def _request_activity_data(self, athlete, filename):
        """
        This method overwrites this method on the GoldenCheetahClient class so
        methods like get_last_activity() return a WorkoutDataFrame instead of a
        regular pandas.DataFrame

        Parameters
        ----------
        athlete : str
        filename : str

        Returns
        -------
        df : WorkoutDataFrame
        """
        df = super(Client, self)._request_activity_data(athlete, filename)
        return WorkoutDataFrame(df)



from collections import OrderedDict

import numpy as np
from scipy.optimize import curve_fit


def two_parameter_non_linear_predict(t, cp, w_prime):
    return cp + w_prime / t


def three_parameter_non_linear_predict(t, cp, w_prime, p_max):
    return w_prime / (t + (w_prime / (p_max - cp))) + cp


def extended_5_3_predict(
    t,
    power_anaerobic_alactic,
    power_anaerobic_decay,
    cp,
    tau_delay,
    cp_delay,
    cp_decay,
    cp_decay_delay,
    tau,
):
    """
    Credits to Damien Grauser. Source:
    https://github.com/GoldenCheetah/GoldenCheetah/blob/master/src/Metrics/ExtendedCriticalPower.cpp
    """
    model = (
        (power_anaerobic_alactic)
        * (1.20 - 0.20 * np.exp(-1 * t))
        * (np.exp(power_anaerobic_decay * t))
    ) + (
        (cp)
        * (1 - np.exp(cp_delay * t))
        * (1 + cp_decay * np.exp(cp_decay_delay / t))
        * (1 - np.exp(tau_delay * t))
        * (1 + tau / t)
    )

    model = np.nan_to_num(model)
    model[model > 2000] = 0
    model[model < 1] = 0

    return model


def extended_7_3_predict(
    t,
    power_anaerobic_alactic,
    power_anaerobic_decay,
    cp,
    tau_delay,
    cp_delay,
    cp_decay,
    cp_decay_delay,
    tau,
):
    """
    Credits to Damien Grauser. Source:
    https://github.com/GoldenCheetah/GoldenCheetah/blob/master/src/Metrics/ExtendedCriticalPower.cpp
    """
    model = (
        (power_anaerobic_alactic)
        * ((np.exp(power_anaerobic_decay)) * (np.power(t, power_anaerobic_alactic)))
    ) + (
        (cp * (1 - np.exp(tau_delay * t)))
        * (1 - np.exp(cp_delay * t))
        * (1 + cp_decay * np.exp(cp_decay_delay / t))
        * (1 + tau / t)
    )

    model = np.nan_to_num(model)
    model[model > 2000] = 0
    model[model < 1] = 0

    return model


def model_fit(x, y, model="extended_5_3"):
    if model == "2_parameter_non_linear":
        predict_func = two_parameter_non_linear_predict
        initial_model_params = OrderedDict(cp=300, w_prime=20000)
    if model == "3_parameter_non_linear":
        predict_func = three_parameter_non_linear_predict
        initial_model_params = OrderedDict(cp=300, w_prime=20000, p_max=1000)
    elif model == "extended_5_3":
        predict_func = extended_5_3_predict
        initial_model_params = OrderedDict(
            power_anaerobic_alactic=811,
            power_anaerobic_decay=-2,
            cp=280,
            cp_delay=-0.9,
            cp_decay=-0.583,
            cp_decay_delay=-180,
            tau=1.208,
            tau_delay=-4.8,
        )
    elif model == "extended_7_3":
        predict_func = extended_7_3_predict
        initial_model_params = OrderedDict(
            power_anaerobic_alactic=811,
            power_anaerobic_decay=-2,
            cp=280,
            cp_delay=-0.9,
            cp_decay=-0.583,
            cp_decay_delay=-180,
            tau=1.208,
            tau_delay=-4.8,
        )

    initial_param_array = np.array(list(initial_model_params.values()))

    model_params, _ = curve_fit(
        f=predict_func, xdata=x, ydata=y, p0=initial_param_array,
    )

    fitted_model_params = dict(zip(initial_model_params.keys(), model_params))

    return fitted_model_params



import numpy as np

from sweat.io.models.exceptions import WorkoutDataFrameValidationException

VALIDATION_COLUMNS = ["time", "power", "heartrate", "cadence", "speed"]

COLUMN_DTYPES = {
    "time": np.int,
    "power": np.int,
    "heartrate": np.int,
    "cadence": np.int,
    "speed": np.int,
}

COLUMN_MIN_VALUE = {"time": 0, "power": 0, "heartrate": 0, "cadence": 0, "speed": 0}

COLUMN_MAX_VALUE = {
    "time": np.inf,
    "power": 3000,
    "heartrate": 300,
    "cadence": 300,
    "speed": 0,
}


class ValidationMixin(object):
    def is_valid(self):
        errors = list()

        errors += self._validate_index()
        errors += self._validate_columns()

        if errors:
            raise WorkoutDataFrameValidationException(errors)
        return True

    def _validate_columns(self):
        errors = list()
        columns = set(VALIDATION_COLUMNS) & set(list(self))

        for c in columns:
            errors += self._validate_series(
                series=self[c],
                dtype=COLUMN_DTYPES[c],
                min_value=COLUMN_MIN_VALUE[c],
                max_value=COLUMN_MAX_VALUE[c],
            )

        return errors

    @classmethod
    def _validate_series(cls, series, dtype, min_value, max_value):
        errors = list()
        if series.dtype != dtype:
            errors.append("Column '{}' is not of dtype '{}'".format(series.name, dtype))
        if series.min() < min_value:
            errors.append("Column '{}' has values < {}".format(series.name, min_value))
        if series.max() > max_value:
            errors.append("Column '{}' has values > {}".format(series.name, max_value))
        return errors

    def _validate_index(self):
        errors = list()

        if not self.index.name == "time":
            errors.append(
                "Index should be set to 'time'. See 'pandas.DataFrame.set_index'"
            )
            return errors

        # @TODO change diff to calculation that's meaningful for time object
        # @TODO also change time dtype and min max
        index_diff = self.index.to_series().diff(1)
        if index_diff.where(index_diff != 1).where(index_diff != np.nan).count():
            errors.append("Sample rate is not (consistent) 1Hz")

        return errors



class WorkoutDataFrameValidationException(Exception):
    def __init__(self, errors):
        self.message = "Validation failed at:\n{}".format("\n".join(errors))
        super(WorkoutDataFrameValidationException, self).__init__(self.message)



from sweat.hrm import heartrate_models
from sweat.pdm import w_prime_balance
from sweat.metrics import core, power
from .base import BaseWorkoutDataFrame


class WorkoutDataFrame(BaseWorkoutDataFrame):
    """Pandas DataFrame for Workouts

    Columns
    -------
    power : array-like
    heartrate : array-like
    time : array-like
    cadence : array-like
    speed : array-like

    Metadata
    --------
    athlete : Athlete
    """

    _metadata = ["athlete"]

    def compute_mean_max_power(self):
        return core.mean_max(self.power)

    def compute_weighted_average_power(self):
        return core.weighted_average_power(self.power, algorithm="WAP")

    def compute_power_per_kg(self):
        return power.wpk(self.power, self.athlete.weight)

    def compute_w_prime_balance(self, algorithm=None, *args, **kwargs):
        return w_prime_balance.w_prime_balance(
            self.power,
            self.athlete.cp,
            self.athlete.w_prime,
            algorithm,
            *args,
            **kwargs
        )

    def compute_mean_max_bests(self, duration, number):
        return core.multiple_best_intervals(self.power, duration, number)

    def compute_heartrate_model(self):
        return heartrate_models.heartrate_model(self.heartrate, self.power)


class Athlete:
    """Athlete object for WorkoutDataFrame"""

    def __init__(
        self,
        name=None,
        sex=None,
        weight=None,
        dob=None,
        ftp=None,
        cp=None,
        w_prime=None,
    ):
        """

        Parameters
        ----------
        name : str, optional
        sex : srt, optional
        weight : number, optional
        dob : str, "YYYY-MM-DD", optional
        ftp : number, optional
        cp : number, optional
        w_prime : number, optional
        """
        self.name = name
        self.sex = sex
        self.weight = weight
        self.dob = dob
        self.ftp = ftp
        self.cp = cp
        self.w_prime = w_prime



import pathlib

from fitparse import FitFile
import numpy as np
import pandas as pd


def read_fit(fpath):
    """
    This method uses the Python fitparse library to load a FIT file into a Pandas DataFrame.
    It is tested with a Garmin FIT file but will probably work with other FIT files too.
    Columns names are translated to sweat terminology (e.g. "heart_rate" > "heartrate").

    Parameters
    ----------
    fpath : str, file-like or Path object

    Returns
    -------
    fit_df : Pandas DataFrame
    """

    if isinstance(fpath, pathlib.PurePath):
        fpath = fpath.as_posix()

    fitfile = FitFile(fpath)

    records = []
    for record in fitfile.get_messages("record"):
        records.append(record.get_values())

    fit_df = pd.DataFrame(records)

    fit_df = fit_df.rename(columns={"heart_rate": "heartrate"})

    fit_df.index = (fit_df["timestamp"] - fit_df["timestamp"][0]) / np.timedelta64(
        1, "s"
    )
    fit_df.index = fit_df.index.astype(int)

    return fit_df



import pandas as pd

from sweat.io.models.mixins import ValidationMixin


class BaseWorkoutDataFrame(pd.DataFrame, ValidationMixin):
    _metadata = ["athlete"]

    @property
    def _constructor(self):
        return self.__class__  # A less hacky way would be appreciated...
